#!/usr/bin/env python
# This script has been automatically generated by AbiPy.
from __future__ import annotations

import numpy as np
import pandas as pd

if False:
    import seaborn as sns
    sns.set(context="paper", style='darkgrid', palette='deep',
            font='sans-serif', font_scale=0.8, color_codes=False, rc=None)


from abipy.ml.relax_scanner import RelaxScannerAnalyzer
rsa = RelaxScannerAnalyzer.from_topdir(".")
print_dataframe(rsa.df)
#rsa.histplot()

rsa.pairs_enediff_dist(ene_diff=1e-3, dist_tol=3.0)
