"""This module contains a list of default tags for the AbinitInput object"""

GROUND_STATE = "ground_state"
SCF = "scf"
RELAX = "relax"
ION_RELAX = "ion_relax"
IONCELL_RELAX = "ioncell_relax"
NSCF = "nscf"
BANDS = "bands"
DOS = "dos"
MOLECULAR_DYNAMICS = "molecular_dynamics"
DFPT = "dfpt"
PHONON = "phonon"
DDK = "ddk"
DDE = "dde"
BEC = "bec"
PH_Q_PERT = "ph_q_pert"
STRAIN = "strain"
MANY_BODY = "many_body"
SIGMA = "sigma"
SCREENING = "screening"
BSE = "bse"
HYBRID = "hybrid"
DTE = "dte"
