"""
This file contains lists of structures that can be used to test flows for HTC.
Preferably one should test the scripts on all lists.
"""

# some simple semiconductors Si, C (diamond), LiF, MgO, GaN
simple_semiconductors = ['mp-149', 'mp-66', 'mp-830', 'mp-1138', 'mp-1265']

# simple metals: Na, Mg, Al
simple_metals = ['mp-10172', 'mp-153', 'mp-134']
